from __future__ import annotations

from textwrap import indent
from typing import Generic, TypeVar, Callable, Optional, Union, List, Dict, Any
from struct import pack, unpack, calcsize

from memhax.constants import CONFIG, memory

T = TypeVar('T')
R = TypeVar('R')


class Proxy(Generic[T]):
    def __init__(self, addr: int, getter: Optional[Callable[[], T]] = None, setter: Optional[Callable[[Union[T, Proxy[T]]], None]] = None, size: int = 0):
        def _throwNotImplemented(*args):
            raise NotImplementedError("No getter or setter defined for this proxy")

        self.address = addr
        self._getter = getter or _throwNotImplemented
        self._setter = setter or _throwNotImplemented
        self._offset = 0
        self._size = size
        self._fields: Dict[str, Proxy[Any]] = {}

    def _align(self, _type: Callable[[int], Proxy[Any]]) -> None:
        dummy = _type(0)
        alignment = dummy.alignment()
        rem = (self.address + self._offset) % alignment
        if rem != 0:
            self._offset += rem

    def _property(self, name: str, _type: Callable[[int], Proxy[R]]) -> Proxy[R]:
        self._align(_type)

        self._fields[name] = obj = _type(self.address + self._offset)
        self._offset += obj.sizeof()
        return obj

    def __getattr__(self, item) -> Proxy[Any]:
        return self._fields[item]

    def __call__(self, arg: Optional[Union[T, Proxy[T]]] = None) -> T:
        if arg is not None:
            self._setter(arg)
            return arg
        return self._getter()

    def alignment(self) -> int:
        raise NotImplementedError(self.__class__.__name__)

    def sizeof(self) -> int:
        return self._size + self._offset

    def offsetof(self, field: Proxy[Any]) -> int:
        return field.address - self.address

    def offsetafter(self, field: Proxy[Any]) -> int:
        return field.address + field.sizeof() - self.address

    # === Properties ===

    def long(self, name: str) -> Proxy[int]:
        from memhax.native import Long
        return self._property(name, Long)

    def ulong(self, name: str) -> Proxy[int]:
        from memhax.native import UnsignedLong
        return self._property(name, UnsignedLong)

    def int(self, name: str) -> Proxy[int]:
        from memhax.native import Integer
        return self._property(name, Integer)

    def uint(self, name: str) -> Proxy[int]:
        from memhax.native import UnsignedInteger
        return self._property(name, UnsignedInteger)

    def short(self, name: str) -> Proxy[int]:
        from memhax.native import Short
        return self._property(name, Short)

    def ushort(self, name: str) -> Proxy[int]:
        from memhax.native import UnsignedShort
        return self._property(name, UnsignedShort)

    def byte(self, name: str) -> Proxy[int]:
        from memhax.native import Byte
        return self._property(name, Byte)

    def ubyte(self, name: str) -> Proxy[int]:
        from memhax.native import UnsignedByte
        return self._property(name, UnsignedByte)

    def char(self, name: str) -> Proxy[str]:
        from memhax.native import Char
        return self._property(name, Char)

    def float(self, name: str) -> Proxy[float]:
        from memhax.native import Float
        return self._property(name, Float)

    def double(self, name: str) -> Proxy[float]:
        from memhax.native import Double
        return self._property(name, Double)

    def boolean(self, name: str) -> Proxy[bool]:
        from memhax.native import Boolean
        return self._property(name, Boolean)

    def size_t(self, name: str) -> Proxy[int]:
        from memhax.native import Py_size_t
        return self._property(name, Py_size_t)

    def ssize_t(self, name: str) -> Proxy[int]:
        from memhax.native import Py_ssize_t
        return self._property(name, Py_ssize_t)

    def struct(self, name: str, _type: Callable[[int], Struct[R]]) -> Struct[R]:
        self._align(_type)

        item = _type(self.address + self._offset)
        self._offset += item.sizeof()
        self._fields[name] = item
        return item

    def pointer(self, name: str, from_addr: Optional[Callable[[int], Proxy[R]]] = None) -> Proxy[R]:
        from memhax.native import Pointer
        obj = self._property(name, lambda addr: Pointer(addr, from_addr))
        self._fields[name] = obj
        return obj

    def array(self, name: str, _type: Callable[[int], Proxy[R]], length: Union[int, Callable[[], int]]) -> Proxy[R]:
        from memhax.native import StaticSizeArray, DynamicSizeArray
        _array_type = StaticSizeArray if isinstance(length, int) else DynamicSizeArray
        self._align(lambda addr: _array_type(addr, _type, length))

        obj = _array_type(self.address + self._offset, _type, length)
        self._offset += obj.sizeof()
        self._fields[name] = obj
        return obj

    def type_str(self) -> str:
        raise NotImplementedError(self.__class__.__name__)

    def rich_str(self, objs: List[int] = None) -> str:
        if objs is None:
            objs = []
        if CONFIG["pretty_print"]:
            body = "\n" + ",\n".join(indent(f"{k}={v.rich_str(objs)}", "    ") for k, v in sorted(self._fields.items(), key=lambda x: self.offsetof(x[1]))) + "\n"
        else:
            body = ', '.join(f'{k}={v.rich_str(objs[:])}' for k, v in sorted(self._fields.items(), key=lambda x: self.offsetof(x[1])))
        return f"{self.__class__.__name__}({body})"

    def __repr__(self) -> str:
        return self.rich_str()


class Struct(Proxy[None]):
    def __init__(self, addr: int):
        super().__init__(addr, lambda: None, lambda v: None, 0)

    def alignment(self) -> int:
        return min(self._fields.values(), key=lambda x: self.offsetof(x)).alignment()

    def _reinterpret(self, addr: int) -> Any:
        dummy = (1,)
        from memhax.cpython.collections import PyTupleObject
        _tup_repr = PyTupleObject(id(dummy))
        _ptr_arr = _tup_repr.ob_item()
        _ptr_arr[0](addr)
        return dummy[0]

    def type_str(self) -> str:
        return self.__class__.__name__

    def struct_str(self) -> str:
        pairs = [(field.type_str(), name) for (name, field) in sorted(self._fields.items(), key=lambda x: self.offsetof(x[1]))]
        longest = max(len(_type) for _type, name in pairs)
        fields = '\n'.join(f"    {field.type_str():<{longest}} {name};" for (name, field) in self._fields.items())
        return f"struct {self.__class__.__name__} {{\n{fields}\n}}"

    def __call__(self, arg: Optional[Union[T, Proxy[T]]] = None) -> T:
        if arg is not None:
            raise NotImplementedError
        return self._reinterpret(self.address)


class StructProxy(Proxy[T]):
    def __init__(self, addr: int, fmt: str):
        super().__init__(addr, self._get, self._set, calcsize(fmt))
        self._fmt = fmt

    def alignment(self) -> int:
        return calcsize(self._fmt)

    def rich_str(self, objs: List[int] = None) -> str:
        return repr(self._get())

    def _get(self) -> T:
        with memory(self.address) as mem:
            return unpack(self._fmt, mem.read(self.sizeof()))[0]

    def _set(self, value: Union[T, Proxy[T]]) -> None:
        with memory(self.address) as mem:
            try:
                data = pack(self._fmt, value)
            except TypeError:
                data = pack(self._fmt, value())
            mem.write(data)



from memhax.native import NullTerminatedString
from memhax.proxy import Struct


class PyMethodDef(Struct):
    def __init__(self, addr: int):
        super().__init__(addr)
        self.pointer("ml_name", NullTerminatedString)
        self.pointer("ml_meth")
        self.int("ml_flags")
        self.pointer("ml_doc", NullTerminatedString)


class PyMemberDef(Struct):
    def __init__(self, addr: int):
        super().__init__(addr)
        self.pointer("name", NullTerminatedString)
        self.int("type")
        self.ssize_t("offset")
        self.int("flags")
        self.pointer("doc", NullTerminatedString)


class PyGetSetDef(Struct):
    def __init__(self, addr: int):
        super().__init__(addr)
        self.pointer("name", NullTerminatedString)
        self.pointer("get")
        self.pointer("set")
        self.pointer("doc", NullTerminatedString)
        self.pointer("closure")



from ctypes import c_wchar, sizeof
from struct import unpack

from memhax.constants import memory
from memhax.cpython.object import PyObject, PyVarObject
from memhax.native import Byte, UnsignedInteger


class PyASCIIObject(PyObject):
    def __init__(self, addr: int):
        super().__init__(addr)
        self.ssize_t("length")
        self.ssize_t("hash")
        self.uint("state")
        self.pointer("wstr")

    def _string(self, interned, kind, compact, ascii, ready):
        if not (kind == 1 and compact and ascii and ready):
            raise ValueError("Not a PyASCIIObject valid string:", interned, kind, compact, ascii, ready)
        addr = self.address + self.offsetafter(self.wstr)
        with memory(addr) as mem:
            data = mem.read(self.length())
            return data.decode("ascii")

    def value(self) -> str:
        state = self.state()
        interned = state & 0b11
        kind = (state >> 2) & 0b111
        compact = (state >> 5) & 0b1
        ascii = (state >> 6) & 0b1
        ready = (state >> 7) & 0b1
        return self._string(interned, kind, compact, ascii, ready)


class PyCompactUnicodeObject(PyASCIIObject):
    def __init__(self, addr: int):
        super().__init__(addr)
        self.ssize_t("utf8_length")
        self.pointer("utf8")
        self.ssize_t("wstr_length")

    def _string(self, interned, kind, compact, ascii, ready):
        if kind == 1 and compact and ascii and ready:
            return super()._string(interned, kind, compact, ascii, ready)
        if kind != 0 and compact and not ascii and ready:
            addr = self.address + self.offsetafter(self.wstr_length)
            with memory(addr) as mem:
                data = mem.read(self.length() * kind)
                fmt = {
                    1: "utf-8",
                    2: "utf-16",
                    4: "utf-32"
                }
                return data.decode(fmt[kind])
        raise ValueError("Not a PyCompactUnicodeObject valid string:", interned, kind, compact, ascii, ready)


class PyUnicodeObject(PyCompactUnicodeObject):
    def __init__(self, addr: int):
        super().__init__(addr)
        self.pointer("data")

    def _string(self, interned, kind, compact, ascii, ready):
        if kind == 1 and compact and ascii and ready:
            return super()._string(interned, kind, compact, ascii, ready)  # PyASCIIObject
        if kind != 0 and compact and not ascii and ready:
            return super()._string(interned, kind, compact, ascii, ready)  # PyCompactUnicodeObject
        if kind == 0 and not compact and not ascii and not ready:
            with memory(self.wstr()) as mem:
                data = mem.read(self.wstr_length() * sizeof(c_wchar))  # Depends on ctypes :(
                fmt = {
                    2: "utf-16",
                    4: "utf-32"
                }
                return data.decode(fmt[sizeof(c_wchar)])
        elif kind != 0 and not compact and ready:
            with memory(self.data()) as mem:
                data = mem.read(self.length() * kind)
                fmt = {
                    1: "utf-8",
                    2: "utf-16",
                    4: "utf-32"
                }
                return data.decode(fmt[kind])
        raise ValueError("Not a PyUnicodeObject valid string:", interned, kind, compact, ascii, ready)


class PyBytesObject(PyVarObject):
    def __init__(self, addr: int):
        super().__init__(addr)
        self.ssize_t("ob_shash")
        self.array("ob_sval", Byte, self.ob_size)


class PyLongObject(PyVarObject):
    def __init__(self, addr: int):
        super().__init__(addr)
        self.array("ob_digit", UnsignedInteger, lambda: abs(self.ob_size()))


class PyFloatObject(PyObject):
    def __init__(self, addr: int):
        super().__init__(addr)
        self.double("ob_fval")



import atexit
from contextlib import contextmanager
from typing import Optional

_MEMORY = open("/proc/self/mem", "r+b")
CONFIG = {
    "expand_types": False,
    "hide_pointers": True,
    "pretty_print": True,
}
__all__ = ("CONFIG", "memory")


@contextmanager
def memory(address: Optional[int] = None):
    tmp = _MEMORY.tell()
    if address is not None:
        _MEMORY.seek(address)
    yield _MEMORY
    _MEMORY.seek(tmp)


@atexit.register
def __close_memory():
    _MEMORY.close()



from memhax.cpython.object import PyVarObject, PyObject
from memhax.native import Pointer, DynamicSizeArray, Char
from memhax.proxy import Struct


class PyTupleObject(PyVarObject):
    def __init__(self, addr: int):
        super().__init__(addr)
        self.array("ob_item", lambda _addr: Pointer(_addr, PyObject), self.ob_size)


class PyListObject(PyVarObject):
    def __init__(self, addr: int):
        super().__init__(addr)
        self.pointer("ob_item", lambda _addr: DynamicSizeArray(_addr, lambda _addr2: Pointer(_addr2, PyObject), self.allocated))
        self.ssize_t("allocated")


class PyDictObject(PyObject):
    def __init__(self, addr: int):
        super().__init__(addr)
        self.ssize_t("ma_used")
        self.uint("ma_version_tag_1")
        self.uint("ma_version_tag_2")
        self.pointer("ma_keys", PyDictKeysObject)
        self.pointer("ma_values", lambda _addr: DynamicSizeArray(_addr, lambda _addr2: Pointer(_addr2, PyObject), self.ma_used))


class PyDictKeysObject(Struct):
    def __init__(self, addr: int):
        super().__init__(addr)
        self.ssize_t("dk_refcnt")
        self.ssize_t("dk_size")
        self.pointer("dk_lookup")
        self.ssize_t("dk_usable")
        self.ssize_t("dk_nentries")
        # TODO: dk_entries

        def get_indices_size() -> int:
            size = self.dk_size()
            if size <= 0xff:
                return 1
            elif size <= 0xffff:
                return 2
            elif size <= 0xffffffff:
                return 4
            else:
                return 8

        self.array("dk_indices", Char, get_indices_size)
        # TODO: dk_entries?


class PySetObject(PyObject):
    def __init__(self, addr: int):
        super().__init__(addr)
        self.ssize_t("fill")
        self.ssize_t("used")
        self.ssize_t("mask")
        self.pointer("table", lambda _addr: DynamicSizeArray(_addr, _PySetEntry, self.fill))
        self.ssize_t("hash")
        self.ssize_t("finger")
        self.array("smalltable", _PySetEntry, 8)
        self.pointer("weakreflist", PyObject)


class _PySetEntry(Struct):
    def __init__(self, addr: int):
        super().__init__(addr)
        self.pointer("key", PyObject)
        self.ssize_t("hash")



from memhax.native import NullTerminatedString
from memhax.proxy import Struct


class PyAsyncMethods(Struct):
    def __init__(self, addr: int):
        super().__init__(addr)
        self.pointer("am_await")
        self.pointer("am_aiter")
        self.pointer("am_anext")


class PyNumberMethods(Struct):
    def __init__(self, addr: int):
        super().__init__(addr)
        self.pointer("nb_add")
        self.pointer("nb_subtract")
        self.pointer("nb_multiply")
        self.pointer("nb_remainder")
        self.pointer("nb_divmod")
        self.pointer("nb_power")
        self.pointer("nb_negative")
        self.pointer("nb_positive")
        self.pointer("nb_absolute")
        self.pointer("nb_bool")
        self.pointer("nb_invert")
        self.pointer("nb_lshift")
        self.pointer("nb_rshift")
        self.pointer("nb_and")
        self.pointer("nb_xor")
        self.pointer("nb_or")
        self.pointer("nb_int")
        self.pointer("nb_reserved")
        self.pointer("nb_float")
        self.pointer("nb_inplace_add")
        self.pointer("nb_inplace_subtract")
        self.pointer("nb_inplace_multiply")
        self.pointer("nb_inplace_remainder")
        self.pointer("nb_inplace_power")
        self.pointer("nb_inplace_lshift")
        self.pointer("nb_inplace_rshift")
        self.pointer("nb_inplace_and")
        self.pointer("nb_inplace_xor")
        self.pointer("nb_inplace_or")
        self.pointer("nb_floor_divide")
        self.pointer("nb_true_divide")
        self.pointer("nb_inplace_floor_divide")
        self.pointer("nb_inplace_true_divide")
        self.pointer("nb_index")
        self.pointer("nb_matrix_multiply")
        self.pointer("nb_inplace_matrix_multiply")


class PySequenceMethods(Struct):
    def __init__(self, addr: int):
        super().__init__(addr)
        self.pointer("sq_length")
        self.pointer("sq_concat")
        self.pointer("sq_repeat")
        self.pointer("sq_item")
        self.pointer("was_sq_slice")
        self.pointer("sq_ass_item")
        self.pointer("was_sq_ass_slice")
        self.pointer("sq_contains")
        self.pointer("sq_inplace_concat")
        self.pointer("sq_inplace_repeat")


class PyMappingMethods(Struct):
    def __init__(self, addr: int):
        super().__init__(addr)
        self.pointer("mp_length")
        self.pointer("mp_subscript")
        self.pointer("mp_ass_subscript")


class PyBufferProcs(Struct):
    def __init__(self, addr: int):
        super().__init__(addr)
        self.pointer("bf_getbuffer")
        self.pointer("bf_releasebuffer")



from typing import List

from memhax.constants import CONFIG
from memhax.cpython.methods import PyAsyncMethods, PyNumberMethods, PySequenceMethods, PyMappingMethods, PyBufferProcs
from memhax.cpython.type import PyMemberDef, PyGetSetDef, PyMethodDef
from memhax.native import NullTerminatedString, NullTerminatedArray
from memhax.proxy import Struct


class PyObject(Struct):
    def __init__(self, addr: int):
        super().__init__(addr)
        self.ssize_t("ob_refcnt")
        self.pointer("ob_type", PyTypeObject)

    def incref(self) -> int:
        return self.ob_refcnt(self.ob_refcnt() + 1)

    def decref(self) -> int:
        return self.ob_refcnt(self.ob_refcnt() - 1)


class PyVarObject(PyObject):
    def __init__(self, addr: int):
        super().__init__(addr)
        self.ssize_t("ob_size")


class PyTypeObject(PyVarObject):
    def __init__(self, addr: int):
        super().__init__(addr)
        self.pointer("tp_name", NullTerminatedString)
        self.ssize_t("tp_basicsize")
        self.ssize_t("tp_itemsize")
        self.pointer("tp_dealloc")
        self.ssize_t("tp_vectorcall_offset")
        self.pointer("tp_getattr")
        self.pointer("tp_setattr")
        self.pointer("tp_as_async", PyAsyncMethods)
        self.pointer("tp_repr")
        self.pointer("tp_as_number", PyNumberMethods)
        self.pointer("tp_as_sequence", PySequenceMethods)
        self.pointer("tp_as_mapping", PyMappingMethods)
        self.pointer("tp_hash")
        self.pointer("tp_call")
        self.pointer("tp_str")
        self.pointer("tp_getattro")
        self.pointer("tp_setattro")
        self.pointer("tp_as_buffer", PyBufferProcs)
        self.ulong("tp_flags")
        self.pointer("tp_doc", NullTerminatedString)
        self.pointer("tp_traverse")
        self.pointer("tp_clear")
        self.pointer("tp_richcompare")
        self.ssize_t("tp_weaklistoffset")
        self.pointer("tp_iter")
        self.pointer("tp_iternext")
        self.pointer("tp_methods", lambda _addr: NullTerminatedArray(_addr, PyMethodDef))
        self.pointer("tp_members", lambda _addr: NullTerminatedArray(_addr, PyMemberDef))
        self.pointer("tp_getset", lambda _addr: NullTerminatedArray(_addr, PyGetSetDef))
        self.pointer("tp_base", PyTypeObject)
        self.pointer("tp_dict", PyObject)
        self.pointer("tp_descr_get")
        self.pointer("tp_descr_set")
        self.ssize_t("tp_dictoffset")
        self.pointer("tp_init")
        self.pointer("tp_alloc")
        self.pointer("tp_new")
        self.pointer("tp_free")
        self.pointer("tp_is_gc")
        self.pointer("tp_bases", PyObject)
        self.pointer("tp_mro", PyObject)
        self.pointer("tp_cache")
        self.pointer("tp_subclasses")
        self.pointer("tp_weaklist")
        self.pointer("tp_del")
        self.uint("tp_version_tag")
        self.pointer("tp_finalize")
        self.pointer("tp_vectorcall")

    def rich_str(self, objs: List[int] = None) -> str:
        if CONFIG["expand_types"]:
            return super().rich_str(objs)
        return f"<type {self.tp_name()}>"



from textwrap import indent
from typing import TypeVar, Callable, List, Union, Optional, Generic

from memhax.constants import CONFIG, memory
from memhax.proxy import Proxy, StructProxy


T = TypeVar('T')


class NullTerminatedString(Proxy[str]):
    def __init__(self, addr: int):
        super().__init__(addr, self._get, self._set)

    def rich_str(self, objs: List[int] = None) -> str:
        return repr(self._get())

    def _get(self) -> str:
        with memory(self.address) as mem:
            string = b""
            while (c := mem.read(1)) != b'\x00':
                string += c
            return string.decode("utf-8")

    def _set(self, value: str) -> None:
        with memory(self.address) as mem:
            mem.write(value.encode("utf-8") + b'\x00')

    def type_str(self):
        return f"char"


class Long(StructProxy[int]):
    def __init__(self, addr: int):
        super().__init__(addr, "q")

    def type_str(self):
        return "long"


class UnsignedLong(StructProxy[int]):
    def __init__(self, addr: int):
        super().__init__(addr, "Q")

    def type_str(self):
        return "unsigned long"


class Integer(StructProxy[int]):
    def __init__(self, addr: int):
        super().__init__(addr, "i")

    def type_str(self):
        return "int"


class UnsignedInteger(StructProxy[int]):
    def __init__(self, addr: int):
        super().__init__(addr, "I")

    def type_str(self):
        return "unsigned int"


class Short(StructProxy[int]):
    def __init__(self, addr: int):
        super().__init__(addr, "h")

    def type_str(self):
        return "short"


class UnsignedShort(StructProxy[int]):
    def __init__(self, addr: int):
        super().__init__(addr, "H")

    def type_str(self):
        return "unsigned short"


class Byte(StructProxy[int]):
    def __init__(self, addr: int):
        super().__init__(addr, "b")

    def type_str(self):
        return "byte"


class UnsignedByte(StructProxy[int]):
    def __init__(self, addr: int):
        super().__init__(addr, "B")

    def type_str(self):
        return "unsigned byte"


class Char(StructProxy[str]):
    def __init__(self, addr: int):
        super().__init__(addr, "c")

    def type_str(self):
        return "char"


class Float(StructProxy[float]):
    def __init__(self, addr: int):
        super().__init__(addr, "f")

    def type_str(self):
        return "float"


class Double(StructProxy[float]):
    def __init__(self, addr: int):
        super().__init__(addr, "d")

    def type_str(self):
        return "double"


class Boolean(StructProxy[bool]):
    def __init__(self, addr: int):
        super().__init__(addr, "?")

    def type_str(self):
        return "boolean"


class Py_size_t(StructProxy[int]):
    def __init__(self, addr: int):
        super().__init__(addr, "N")

    def type_str(self):
        return "Py_size_t"


class Py_ssize_t(StructProxy[int]):
    def __init__(self, addr: int):
        super().__init__(addr, "n")

    def type_str(self):
        return "Py_ssize_t"


class Pointer(StructProxy[Proxy[T]]):
    def __init__(self, addr: int, _type: Optional[Callable[[int], Proxy[T]]] = None):
        super().__init__(addr, "P")
        self._type = _type

    def type_str(self):
        return f"{self._get().type_str()}*"

    def rich_str(self, objs: List[int] = None) -> str:
        if objs is None:
            objs = []
        addr = self.raw()
        if addr == 0:
            return "NULL"
        if addr in objs:
            return "..."
        objs.append(addr)
        if self._type is None:
            return f"{self.__class__.__name__}(0x{addr:x})"
        child = self._get().rich_str(objs[:])
        if CONFIG["hide_pointers"]:
            return child
        return f"*{child}"

    def raw(self) -> int:
        return super()._get()

    def _get(self) -> Proxy[T]:
        if self._type is None:
            return Proxy(self.raw())
        return self._type(super()._get())

    def _set(self, value: Union[int, Proxy[T]]) -> None:
        if isinstance(value, Proxy):
            super()._set(value.address)
        else:
            super()._set(value)

    def __getattr__(self, item):
        return getattr(self._get(), item)


class _ArrayCommon(Generic[T]):
    def _get(self) -> List[Proxy[T]]:
        raise NotImplementedError()

    def alignment(self) -> int:
        return self._get()[0].alignment()

    def type_str(self):
        items = self._get()
        return f"{items[0].type_str()}[{len(items)}]"

    def rich_str(self, objs: List[int] = None) -> str:
        items = self._get()

        if objs is None:
            objs = []
        if CONFIG["pretty_print"]:
            body = "" if not items else "\n" + ",\n".join(indent(item.rich_str(objs[:]), "    ") for item in items) + "\n"
        else:
            body = ', '.join([item.rich_str(objs[:]) for item in items])
        return f"[{body}]"

    def __getitem__(self, item):
        return self._get()[item]

    def __setitem__(self, key, value):
        self._get()[key](value)


class StaticSizeArray(_ArrayCommon[T], StructProxy[List[Proxy[T]]]):
    def __init__(self, addr: int, _type: Callable[[int], Proxy[T]], size: int):
        super().__init__(addr, f"{size}P")
        self._type = _type
        self._items = []

        offset = 0

        def _align():
            nonlocal offset
            dummy = _type(0)
            alignment = dummy.alignment()
            rem = (self.address + offset) % alignment
            if rem != 0:
                offset += rem

        for i in range(size):
            _align()
            item = _type(self.address + offset)
            self._items.append(item)
            offset += item.sizeof()

    def _get(self) -> List[Proxy[T]]:
        return self._items

    def _set(self, value: List[Proxy[T]]) -> None:
        if len(value) != len(self._items):
            raise ValueError("Array size mismatch")

        for old, new in zip(self._items, value):
            old(new)


class DynamicSizeArray(_ArrayCommon[T], Proxy[List[Proxy[T]]]):
    def __init__(self, addr: int, _type: Callable[[int], Proxy[T]], size: Callable[[], int]):
        super().__init__(addr, self._get, self._set)
        self._type = _type
        self._length = size

    def _get(self) -> List[Proxy[T]]:
        items = []
        offset = 0

        def _align():
            nonlocal offset
            dummy = self._type(0)
            alignment = dummy.alignment()
            rem = (self.address + offset) % alignment
            if rem != 0:
                offset += rem

        for i in range(self._length()):
            _align()
            item = self._type(self.address + offset)
            items.append(item)
            offset += item.sizeof()

        return items

    def _set(self, value: List[Proxy[T]]) -> None:
        if len(value) != self._length():
            raise ValueError("Array size mismatch")

        original = self._get()
        for old, new in zip(original, value):
            old(new)


class NullTerminatedArray(_ArrayCommon[T], Proxy[List[Proxy[T]]]):
    def __init__(self, addr: int, _type: Callable[[int], Proxy[T]]):
        super().__init__(addr, self._get, self._set)
        self._type = _type

    def _get(self) -> List[Proxy[T]]:
        items = []
        offset = 0

        def _align():
            nonlocal offset
            dummy = self._type(0)
            alignment = dummy.alignment()
            rem = (self.address + offset) % alignment
            if rem != 0:
                offset += rem

        for i in range(self._length()):
            _align()
            item = self._type(self.address + offset)
            MEMORY.seek(self.address + offset)
            if all(x == b"\0" for x in MEMORY.read(item.sizeof())):
                break
            items.append(item)
            offset += item.sizeof()

        return items

    def _set(self, value: List[Proxy[T]]) -> None:
        original = self._get()
        if len(value) != len(original):
            raise ValueError("Array size mismatch")

        for old, new in zip(original, value):
            old(new)
